"""
Blueprint Domain Population

"""
# System
import logging
import logging.config
import sys
import argparse
from pathlib import Path
import atexit

# Domain Population
from dpop.system import System
from dpop import version

_logpath = Path("domaindb.log")
_progname = 'Blueprint Domain Population'

def clean_up():
    """Normal and exception exit activities"""
    _logpath.unlink(missing_ok=True)

def get_logger():
    """Initiate the logger"""
    log_conf_path = Path(__file__).parent / 'log.conf'  # Logging configuration is in this file
    logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
    return logging.getLogger(__name__)  # Create a logger for this module

# Configure the expected parameters and actions for the argparse module
def parse(cl_input):
    parser = argparse.ArgumentParser(description=_progname)
    parser.add_argument('-s', '--system', action='store',
                        help='Name of the metamodel TclRAL database *.ral file populated with one or more domains')
    parser.add_argument('-c', '--context', action='store',
                        help='Name of the context *.sip file specifying initial populations and states per domain')
    parser.add_argument('-t', '--types', action='store',
                        help='Name of the *.yaml file that maps domain to TclRAL supported data types')
    parser.add_argument('-o', '--output', action='store_true',
                        help='Output each populated database to standard output as *.txt files, on per domain db')
    parser.add_argument('-D', '--debug', action='store_true',
                        help='Debug mode'),
    parser.add_argument('-E', '--examples', action='store_true',
                        help='Create a directory of examples in the current directory')
    parser.add_argument('-L', '--log', action='store_true',
                        help='Generate a diagnostic log file')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Verbose messages -- Tables for both the system database and each domain database will be displayed')
    parser.add_argument('-V', '--version', action='store_true',
                        help='Print the current version')
    return parser.parse_args(cl_input)


def main():
    # Start logging
    logger = get_logger()
    logger.info(f'{_progname} version: {version}')

    # Parse the command line args
    args = parse(sys.argv[1:])

    if args.examples:
        # Copy the entire example directory into the users local dir if it does not already exist
        import shutil
        ex_path = Path(__file__).parent.parent / 'examples'
        local_ex_path = Path.cwd() / 'examples'
        if local_ex_path.exists():
            logger.warning("Examples already exist in the current directory. Delete or move it if you want the latest.")
        else:
            logger.info("Copying example directory into user's local directory")
            shutil.copytree(ex_path, local_ex_path)  # Copy the example directory

    if args.examples:
        # Don't generate any domain dbs
        # Just quit here
        sys.exit(0)

    if args.version:
        # Just print the version and quit
        print(f'{_progname} version: {version}')
        sys.exit(0)

    if not args.log:
        # If no log file is requested, remove the log file before termination
        atexit.register(clean_up)

    # System specified
    if args.system:
        s = System(mmdb_path=Path(args.system), context_path=Path(args.context),
                   types_path=Path(args.types), verbose=args.verbose, output_text=args.output, debug=args.debug)

    logger.info("No problemo")  # We didn't die on an exception, basically
    if args.verbose:
        print("\nNo problemo")


if __name__ == "__main__":
    main()



""" domain_model_db.py -- Build a TclRAL schema for a modeled domain from a populated SM Metamodel """

# System
import logging
import yaml
from collections import defaultdict
from typing import TYPE_CHECKING, Any, NamedTuple
from contextlib import redirect_stdout
from pathlib import Path

if TYPE_CHECKING:
    from dpop.system import System

# Model Integration
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute, Mult

# Domain Populate
from dpop.db_names import mmdb
from dpop.context import Context
from dpop.exceptions import *

_logger = logging.getLogger(__name__)

# Here is a mapping from metamodel multiplcity notation to that used by the target TclRAL tclral
# When interacting with PyRAL we must supply the tclral specific value
mult_tclral = {
    'M': Mult.AT_LEAST_ONE,
    '1': Mult.EXACTLY_ONE,
    'Mc': Mult.ZERO_ONE_OR_MANY,
    '1c': Mult.ZERO_OR_ONE
}

def load_yaml(file_path: Path) -> Any:
    """
    Load the specified YAML file and return the loaded data
    :param file_path:
    :return: Loaded data (type depends on content of the file)
    """
    try:
        with file_path.open("r") as file:
            data = yaml.safe_load(file)
        return data
    except FileNotFoundError:
        _logger.error(f"Error: File not found — {file_path}")
    except PermissionError:
        _logger.error(f"Error: Permission denied — {file_path}")
    except yaml.YAMLError as e:
        _logger.error(f"Error: Invalid YAML format in {file_path}\nDetails: {e}")
    except Exception as e:
        _logger.error(f"Unexpected error reading {file_path}: {e}")
    raise DPOPFileException(f"File [{file_path}] not loaded")


MultipleAssigner = NamedTuple("MultipleAssigner", rnum=str, pclass=str)

class DomainModelDB:
    """
    TclRAL schema for a user domain model extracted from a populated SM metamodel
    """

    def __init__(self, name: str, alias: str, system: 'System'):
        """
        Create the db schema and optionally print it out

        :param name: Name of this domain
        :param alias: Alias for this domain (used for the database name)
        :param system: The system object
        """
        self.system = system
        self.domain = name
        self.alias = alias
        self.ordinal_rnums = None
        self.gen_rnums = None
        self.non_assoc_rnums = None
        self.assoc_rnums = None
        self.user_types = None
        self.context = None

        Database.open_session(name=self.alias)  # User models created in this database

        self.build_class_relvars()
        self.sort_rels()
        self.build_simple_assocs()
        self.build_associative_rels()
        self.build_gen_rels()
        self.populate()
        if self.system.verbose:
            self.display()
        if self.system.output_text:
            self.print()
        self.save()

    def display(self):
        """
        Display the user domain schema on the console
        """
        msg = f"Populated {self.domain} domain model"
        print(f"\n*** {msg} ***")
        Relvar.printall(db=self.alias)
        print(f"\n^^^ {msg} ^^^\n")

    def print(self):
        """
        Print out the user domain schema
        """
        with open(f"{self.alias.lower()}.txt", 'w') as f:
            with redirect_stdout(f):
                Relvar.printall(db=self.alias)

    def populate(self):
        """
        Populate the domain database with initial instances using the supplied context.
        """
        self.context = Context(domaindb=self)

    def save(self):
        """
        Save the domain database as an *.ral file in the current working directory.
        """
        Database.save(db=self.alias, fname=f"{self.alias}.ral")

    def build_gen_rels(self):
        """
        Create referential constraints for each generalization relationship
            name: 'R14'
            superclass_name: 'Subsystem Element'
            super_attrs: ['Label', 'Domain']
            subs: {'Relationship': ['Rnum', 'Domain'], 'Class':['Cnum', 'Domain']}
        """
        for g in self.gen_rnums:
            # Get the name of the superclass
            R = f"Rnum:<{g}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Generalization', restriction=R)
            superclass = result.body[0]['Superclass']

            # Get the superclass referenced identifier attributes
            # This is the same for each subclass attribute reference, so we just take one
            # of them and grab the To Attributes in the reference
            result = Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R)
            one_subclass = result.body[0]['From_class']
            R = f"Rnum:<{g}>, From_class:<{one_subclass}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R)
            superclass_attrs = [ref['To_attribute'].replace(' ', '_') for ref in result.body]

            # Create a dictionary of subclass names and to attrs from each
            # First get a list of subclasses
            R = f"Rnum:<{g}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Subclass', restriction=R)
            subclass_names = [s['Class'] for s in result.body]
            # Now create a dictionary with list of referring attributes keyed by each subclass name
            subclasses = dict()
            for sub_name in subclass_names:
                R = f"Rnum:<{g}>, From_class:<{sub_name}>, Domain:<{self.domain}>"
                result = Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R)
                subclasses[sub_name.replace(' ','_')] = [ref['From_attribute'].replace(' ', '_') for ref in result.body]

            # Create the generalization constraint
            Relvar.create_partition(db=self.alias, name=g, superclass_name=superclass.replace(' ','_'),
                                    super_attrs=superclass_attrs, subs=subclasses)

    def build_associative_rels(self):
        """
        Create referential constraints for each association formalized by an association class
        """
        for a in self.assoc_rnums:
            # Get the name of the association class
            R = f"Rnum:<{a}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Association_Class', restriction=R)
            assoc_class = result.body[0]['Class']

            # Get the mult/cond of each participating class (referenced from the association class)
            result = Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
            ref1_class = result.body[0]['Viewed_class']
            conditional = 'c' if result.body[0]['Conditional'] == 'True' else ''
            ref1_mult = result.body[0]['Multiplicity'] + conditional
            ref2_class = result.body[1]['Viewed_class']
            conditional = 'c' if result.body[1]['Conditional'] == 'True' else ''
            ref2_mult = result.body[1]['Multiplicity'] + conditional

            # Get the from/to attrs for ref1
            R = f"To_class:<{ref1_class}>, Rnum:<{a}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R)
            ref1_from_attrs = [rc['From_attribute'].replace(' ', '_') for rc in result.body]
            ref1_to_attrs = [rc['To_attribute'].replace(' ', '_') for rc in result.body]

            # Get the from/to attrs for ref2
            R = f"To_class:<{ref2_class}>, Rnum:<{a}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R)
            ref2_from_attrs = [rc['From_attribute'].replace(' ', '_') for rc in result.body]
            ref2_to_attrs = [rc['To_attribute'].replace(' ', '_') for rc in result.body]

            Relvar.create_correlation(db=self.alias, name=a, correlation_relvar=assoc_class.replace(' ', '_'),
                                      correl_a_attrs=ref1_from_attrs, a_mult=mult_tclral[ref1_mult],
                                      a_relvar=ref1_class.replace(' ', '_'), a_ref_attrs=ref1_to_attrs,
                                      correl_b_attrs=ref2_from_attrs, b_mult=mult_tclral[ref2_mult],
                                      b_relvar=ref2_class.replace(' ', '_'), b_ref_attrs=ref2_to_attrs,
                                      )

    def build_simple_assocs(self):
        """
        Create referential constraints for each non-associative association
        i.e. Associations not formalized by an association class
        """
        for a in self.non_assoc_rnums:
            R = f"Rnum:<{a}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R)
            from_class = result.body[0]['From_class']
            to_class = result.body[0]['To_class']
            from_attrs = [ref['From_attribute'].replace(" ", "_") for ref in result.body]
            to_attrs = [ref['To_attribute'].replace(" ", "_") for ref in result.body]
            R = f"Viewed_class:<{from_class}>, Rnum:<{a}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
            conditional = 'c' if result.body[0]['Conditional'] == 'True' else ''
            from_mult = result.body[0]['Multiplicity'] + conditional
            R = f"Viewed_class:<{to_class}>, Rnum:<{a}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
            conditional = 'c' if result.body[0]['Conditional'] == 'True' else ''
            to_mult = result.body[0]['Multiplicity'] + conditional
            Relvar.create_association(db=self.alias, name=a,
                                      from_relvar=from_class.replace(" ", "_"), from_attrs=from_attrs, from_mult=mult_tclral[from_mult],
                                      to_relvar=to_class.replace(" ", "_"), to_attrs=to_attrs, to_mult=mult_tclral[to_mult],
                                      )

    def sort_rels(self):
        """
        Sort the domain's rnums into non-associative, associative, and generalization relationships
        """
        # Simple associations
        # Get rnums from Association class
        # Subtract from all rnums in Association
        R = f"Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Association', restriction=R, svar_name='assocs')
        Relation.project(db=mmdb, attributes=('Rnum',), svar_name='assocs')
        Relation.restrict(db=mmdb, relation='Association_Class', restriction=R, svar_name='aclasses')
        result = Relation.project(db=mmdb, attributes=('Rnum',), svar_name='aclasses')
        self.assoc_rnums = [r['Rnum'] for r in result.body]
        result = Relation.subtract(db=mmdb, rname2='aclasses', rname1='assocs')
        self.non_assoc_rnums = [r['Rnum'] for r in result.body]
        result = Relation.restrict(db=mmdb, relation='Generalization', restriction=R)
        self.gen_rnums = [r['Rnum'] for r in result.body]
        result = Relation.restrict(db=mmdb, relation='Ordinal_Relationship', restriction=R)
        self.ordinal_rnums = [r['Rnum'] for r in result.body]

    def build_class_relvars(self):
        """
        Create class relvars
        """
        self.user_types = load_yaml(self.system.types_path)

        R = f"Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Class', restriction=R, svar_name='classes')
        result = Relation.project(db=mmdb, attributes=('Name', 'Cnum'), svar_name='classes')
        classes = result.body
        for c in classes:
            cname = c['Name'].replace(" ", "_")

            # Get the identifiers
            R = f"Class:<{c['Name']}>, Domain:<{self.domain}>"
            Relation.restrict(db=mmdb, relation='Identifier_Attribute', restriction=R, svar_name='idattrs')
            result = Relation.project(db=mmdb, attributes=('Identifier', 'Attribute'), svar_name='idattrs')
            id_attrs = result.body
            # Go get the attributes
            R = f"Class:<{c['Name']}>, Domain:<{self.domain}>"
            Relation.restrict(db=mmdb, relation='Attribute', restriction=R, svar_name='attrs')
            result = Relation.project(db=mmdb, attributes=('Name', 'Scalar'), svar_name='attrs')
            attrs = result.body

            attr_list = [Attribute(name=a['Name'], type=self.user_types[a['Scalar']]) for a in attrs]

            id_dict = defaultdict(list)

            for i in id_attrs:
                key = int(i['Identifier'])  # Convert Identifier to int if needed
                id_dict[key].append(i['Attribute'].replace(" ", "_"))
            ids = dict(id_dict)

            Relvar.create_relvar(db=self.alias, name=cname, attrs=attr_list, ids=ids)



""" System.py -- A Shlaer-Mellor Executable UML System of modeled Domains in a metamodel database """

# System
import logging
from pathlib import Path

# Model Integration
from pyral.relation import Relation

# Domain Populate
from dpop.metamodel_db import MetamodelDB
from dpop.domain_model_db import DomainModelDB
from dpop.db_names import mmdb
from dpop.exceptions import *

_logger = logging.getLogger(__name__)

class System:
    """
    In Shlaer-Mellor Executable UML, a System consists of one or more modeled and non-modeled Domains.

    Here we are focused exclusively on the modeled Domains populated into a metamodel database.

    The System's role here is to output a separate TclRAL database file
    for each domain populated into a metamodel database.
    """

    def __init__(self, mmdb_path: Path, context_path: Path, types_path: Path, output_text: bool,
                 verbose: bool, debug: bool = False):
        """
        Create a domain database for each domain in the populated metamodel and then populate each of these
        domains with instances specified context.

        :param mmdb_path: Path to the populated metamodel *.ral TclRAL
        :param context_path: A *.sip file specifying an instance population for each domain
        :param type_mapping: A *.yaml file specifying a mapping of domain to TclRAL data types
        param verbose: Verbose mode has been set
        :param output_text: If true, each populated domain db is displayed as tables on the console
        :param debug: Debug mode - prints schemas and other info to the console if true
        """
        self.mmdb_path = mmdb_path
        self.context_path = context_path
        self.output_text = output_text
        self.types_path = types_path
        self.verbose = verbose
        self.debug = debug

        # Load a metamodel file populated with a system
        MetamodelDB.load(mmdb_path=self.mmdb_path)
        if self.verbose:
            MetamodelDB.display()
        if self.debug:
            MetamodelDB.print()

        # Set the system name
        # There should be one System instance in the System name from the populated metamodel
        system_r = Relation.restrict(db=mmdb, relation='System')  # System relation with one tuple
        if not system_r.body:
            msg = f"System name not found in the populated metamodel db"
            _logger.exception(msg)
            raise MMDBDataMissing(msg)

        self.name = system_r.body[0]['Name']

        # Create a database schema for each domain
        domain_r = Relation.restrict(db=mmdb, relation='Domain')
        if not domain_r.body:
            msg = f"No domains defined for system in metamodel"
            _logger.exception(msg)
            raise MMDBDataMissing(msg)

        self.domain_dbs = {d['Alias']: DomainModelDB(name=d['Name'], alias=d['Alias'], system=self)
                           for d in domain_r.body}



""" metamodel_db.py -- Loads a metamodel database populated with a system of one or more modeled domains """

# System
import logging
from pathlib import Path
from contextlib import redirect_stdout

# Model Integration
from pyral.database import Database
from pyral.relvar import Relvar

# Model Execution
from dpop.db_names import mmdb

_logger = logging.getLogger(__name__)

class MetamodelDB:

    filename = None

    @classmethod
    def load(cls, mmdb_path: Path):
        """ Let's load and print out the metamodel database """

        cls.filename = mmdb_path

        _logger.info(f"Loading the metamodel database from: [{mmdb_path}]")
        Database.open_session(name=mmdb)
        Database.load(db=mmdb, fname=str(mmdb_path))

    @classmethod
    def print(cls):
        """
        Print out the populated metamodel
        """
        with open("mmdb.txt", 'w') as f:
            with redirect_stdout(f):
                Relvar.printall(db=mmdb)

    @classmethod
    def display(cls):
        # Print out the populated metamodel
        msg = f"Metamodel populated from {cls.filename}"
        print(f"\n*** {msg} ***")
        Relvar.printall(db=mmdb)
        print(f"\n^^^ {msg} ^^^\n")



"""
exceptions.py – Domain Populate exceptions
"""

class DPOPException(Exception):
    """ Top level Domain Population exception """
    pass

class MMDBDataMissing(DPOPException):
    """ Data that should be in the metamodel database is missing """
    pass

class DPOPFileException(DPOPException):
    """ File missing """
    pass



""" starting_context.py -- Populate the schema """

# System
from collections import namedtuple
import logging
from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from dpop.domain_model_db import DomainModelDB

# Model Integration
from sip_parser.parser import SIParser
from pyral.relation import Relation
from pyral.relvar import Relvar
from pyral.transaction import Transaction

# Model Execution
from dpop.db_names import mmdb

AttrRef = NamedTuple('AttrRef', from_attr=str, to_attr=str, to_class=str, alias=str)
MultipleAssignerInitialState = NamedTuple('MultipleAssignerInitialState', pclass=str, state=str)

_logger = logging.getLogger(__name__)

tcl_to_python = { 'string': str, 'boolean': bool, 'double': float, 'int': int }

pop_scenario = 'pop'  # Name of transaction that loads the schema

class Context:

    def __init__(self, domaindb: 'DomainModelDB'):
        """
        We see that there is an R1 ref.  We need to find the target attributes and class
        The metamodel gives us Shaft.Bank -> Bank.Name

        We proceed for each instance of Shaft taking the R1 ref (L, P, or F)
        And go find the Bank population
        We find the L instance
        Now we grab the Name value and add it to our Shaft table by adding a key.
        And, at this point, we are building the relation.create command
        When we get all the values, we commit and move on to the next instance

        :param sip_file:  The path to the *.sip file providing the intial instance population
        :param domain:  The subject matter domain being populated
        :param dbtypes: The actual TclRAL db_types used to represent user model db_types
        """
        self.domaindb = domaindb
        self.lifecycle_istates: dict[str, str] = {}
        self.ma_istates: dict[str, MultipleAssignerInitialState] = {}

        sip_file = self.domaindb.system.context_path

        # Parse the starting_context's initial population file (*.sip file)
        _logger.info(f"Parsing sip: [{sip_file}]")
        parse_result = SIParser.parse_file(file_input=sip_file, debug=False)
        self.name = parse_result.name  # Name of this starting_context
        self.pop = parse_result.classes  # Dictionary of initial instance populations keyed by class name
        self.relations = dict()  # The set of relations keyed by relvar name ready for insertion into the user db

        _logger.info(f"Sip parsed, loading initial instance population into user db")
        # Process each class (c) and its initial instance specification (i)
        for class_name, i_spec in self.pop.items():
            expanded_header = []  # A list of attributes with any references expanded to from_attributes
            instance_tuples = []  # The expanded instance tuples including values for the from_attributes
            ref_path = dict()  # Each attribute reference and data required to resolve it
            for col_position, col in enumerate(i_spec.header):
                # Each column in the parsed header for this class includes a sequence of attributes and
                # optional references. A reference points to some class via a relationship rnum
                if isinstance(col, str):  # Attributes are just string names
                    # The column is an attribute, just add it directly to the expanded header
                    expanded_header.append(col)
                elif isinstance(col, list):  # It must be a list of dictionaries describing a reference
                    for ref in col:
                        # Since an attribute may participate in more than one relationship
                        # there may be multiple references in the same column, for example:
                        #    R38>Bank Level, R5>Bank
                        # is parsed into two components
                        rnum = ref['rnum']  # The rnum on this reference
                        to_class = ref['to class']  # Refering to some target attribute in this class
                        # Look up the attribute reference(s) in the metamodel
                        R = (f"Rnum:<{rnum}>, From_class:<{class_name}>, To_class:<{to_class}>, "
                             f"Domain:<{self.domaindb.domain}>")
                        Relation.restrict(db=mmdb, relation='Attribute_Reference', restriction=R, svar_name='ra')
                        result = Relation.project(db=mmdb, attributes=('From_attribute', 'To_attribute'),
                                                  svar_name='ra')
                        if not result.body:
                            msg = f"Initial instance ref expansion: No attribute references defined for{R}"
                            _logger.exception(msg)
                            raise MXInitialInstanceReferenceException(msg)
                        # We already know the rnum, from class (class_name) and to class, so we just need a projection
                        # on the local attribute and where the attribute in the target class
                        for attr_ref in result.body:
                            # A reference can consist of multiple attributes, so we process each one
                            from_attr = attr_ref['From_attribute']
                            to_attr = attr_ref['To_attribute']
                            # Add a dictionary entry so that we can look up referenced values
                            ref_path[len(expanded_header)] = AttrRef(from_attr=from_attr, to_attr=to_attr,
                                                                     to_class=to_class, alias=col_position)
                            # We key the path to the position of the from_attr in the expanded header
                            # Add the from attribute to our expanded header
                            if from_attr not in expanded_header:
                                # It might already be there if the same attribute participates in more than one
                                # relationship. If so, we only need one value anyway, so add the from attr only
                                # if it isn't already in the header.
                                expanded_header.append(from_attr)
                else:
                    msg = f"Unrecognized column format in initial instance parse: [{col}]"
                    _logger.exception(msg)
                    raise MXException(msg)

            # Now that the relation header for our instance population is created, we need to fill in the relation
            # body (the actual instance values corresponding to each attribute in the expanded header)
            for irow in i_spec.population:
                # save any initial states for classes and multiple assigners
                # TODO: Support single assigners (after SIP support added)
                for s in irow['initial_state']:
                    if len(s) == 1:
                        # save initial state for this class
                        self.lifecycle_istates[class_name] = s[0]
                    if len(s) == 2:
                        # index by rnum and save partitioning class and initial state
                        self.ma_istates[s[0]] = MultipleAssignerInitialState(pclass=class_name, state=s[1])

                # For each instance row under the class header parsed from the init file
                irow_col = 0  # Initial column position in the irow
                row_dict = dict()  # Each value for an instance keyed by attribute name in the expanded header
                in_ref = False  # We are not currently expanding a reference
                for index, attr in enumerate(expanded_header):
                    # We walk through the values matching the attribute order, so we remember the attr ordering
                    if attr in i_spec.header:
                        in_ref = False  # Not processing a reference
                        # If there is no matching key for the attribute in the ref_path, it means that
                        # this was not a reference that got expanded.  So we simply assign the value
                        # from the parsed population to the corresponding attribute in the expanded header
                        attr_value = irow['row'][irow_col]  # This will be a string
                        # Cast it to the appropriate TclRAL type for db insertion
                        cast_attr_value = self.cast_to_dbtype(attr_name=attr, attr_class=class_name, value=attr_value)
                        row_dict[attr.replace(' ', '_')] = cast_attr_value  # We can't have spaces in relvar names
                        irow_col = irow_col + 1  # Increment column position
                    else:
                        if not in_ref:
                            # We are beginning to process a reference, so we want to advance the position
                            # just once (not for every referential attribute)
                            irow_col = irow_col + 1
                        in_ref = True  # This keeps us from incrementing the counter until we're done
                        # The attribute was expanded from a reference and we need to obtain its value from
                        # an attribute in the target class for the instance matching the referenced alias
                        # So first we grab the alias associated with this instance's reference. It tells us which
                        # named instance in the target class to reference. For example this row from Shaft:
                        #    { [S4] [true] @P }
                        # has a third column value with the 'P' alias as designated by the @ character
                        # There is a matching row in the Bank population here:
                        #     P { [Penthouse] [4.0] [2.0] [25] [7.0] [9.0] }
                        #

                        # We want the reference path keyed to the current position in the expanded header
                        ref = ref_path[index]

                        alias = irow['row'][ref.alias]['ref to']  # The alias 'P' in the above example
                        # Get the population of the referenced class
                        target_pop = self.pop[ref.to_class]
                        # Get index of referenced value
                        to_attr_index = target_pop.header.index(ref.to_attr)
                        # Now search through the target population looking for the instance named by the alias
                        referenced_i = [i['row'] for i in target_pop.population if i['alias'] == alias][0]
                        # And then grab the value in that row corresponding to the to_attr_index
                        ref_value = referenced_i[to_attr_index]
                        # And now add the key value pair of referencing attr and target value to our row of values
                        cast_ref_value = self.cast_to_dbtype(attr_name=attr, attr_class=class_name, value=ref_value)
                        row_dict[ref.from_attr] = cast_ref_value
                instance_tuples.append(row_dict)  # Add the completed row of attr values to our relation

            # Now we are ready to create the structure we need to insert into the class relvar in the user db
            expanded_header = [a.replace(' ', '_') for a in expanded_header]  # Replace spaces from any attribute name
            class_tuple_type_name = f"{class_name.replace(' ', '_')}_i"
            ClassTupleType = namedtuple(class_tuple_type_name, expanded_header)
            table = []
            for inst in instance_tuples:
                dvalues = [inst[name] for name in expanded_header]
                drow = ClassTupleType(*dvalues)
                table.append(drow)
            self.relations[class_name] = table
        self.insert()

    def cast_to_dbtype(self, attr_name: str, attr_class: str, value: str) -> int | str | float | bool:
        """
        Casts a string value to a python type (int, str, ...) that correponds to a TclRAL type that
        is used to represent the Scalar defined in the user model.

        Looks up the Scalar associated with the supplied attribute. This is the type specified in the user
        model such as 'Bank Name', 'Duration', etc.

        Consults the user model Scalar -> TclRAL type mapping to determine which type to use in the user db.
        These are low level system db_types that TclRAL supports like 'string', 'int', 'boolean', etc.

        :param attr_name: Name of the user model attribute
        :param attr_class: Name of the attribute's class
        :param value: The value to be cast
        :return: The TclRAL type used to represent the Scalar
        """
        # Look up the user type in the populated metamodel

        R = f"Class:<{attr_class}>, Name:<{attr_name}>, Domain:<{self.domaindb.domain}>"
        result = Relation.restrict(db=mmdb, relation='Attribute', restriction=R)
        if not result.body:
            msg = f"No Scalar found in populated metamodel for attribute [{self.domaindb.domain}:{attr_class}.{attr_name}]"
            _logger.exception(msg)
            raise MXScalarException(msg)
        scalar = result.body[0]['Scalar']
        dbtype = self.domaindb.user_types[scalar]
        # Now cast using corresponding python type
        # Boolean is a special case as it does not provide a string to bool casting function
        if dbtype == 'boolean':
            python_value = True if value.strip().lower() == 'true' else False
        else:
            python_value = tcl_to_python[dbtype](value)
        return python_value

    def insert(self):
        """
        Insert relations in the user database
        """
        Transaction.open(db=self.domaindb.alias, name=pop_scenario)
        for relation, population in self.relations.items():
            Relvar.insert(db=self.domaindb.alias, tr=pop_scenario, relvar=relation.replace(' ', '_'), tuples=population)
        Transaction.execute(db=self.domaindb.alias, name=pop_scenario)

